#coding: utf-8


from django.conf import settings
from django.db import transaction, connections, DEFAULT_DB_ALIAS
from django.test.simple import DjangoTestSuiteRunner, reorder_suite
from django.test import TestCase
from django.test.testcases import disable_transaction_methods, restore_transaction_methods

from bdd_testcase import DjangoBDDTestCase
from wellbehaved.behave_runner import CustomBehaveRunner


class ExistingDBRunner(DjangoTestSuiteRunner):
    def run_tests(self, test_labels, extra_tests=None, **kwargs):
        self.use_existing_db = getattr(settings, 'WELLBEHAVED_USE_EXISTING_DB', False)
        self.setup_test_environment()
        suite = self.build_suite(test_labels, extra_tests)

        if not self.use_existing_db:
            old_config = self.setup_databases()
            result = self.run_suite(suite)
            self.teardown_databases(old_config)
        else:
            result = self.run_suite(suite)

        return self.suite_result(suite, result)


class RollbackBehaveRunner(CustomBehaveRunner):

    def create_savepoint_before_all(self, context):
        u'''
        Обработчик, срабатывающий перед началом выполнения feature.

        Так как behave применяет содержимое Background (Предыстория)
        перед каждым сценарием, мы создаем точку отката транзакции
        для возврата базы в изначальное состояние после отработки
        сценария.

        :param context: Контекстная информация текущего шага.
        '''
        context.__initial_savepoint = transaction.savepoint()

    def restore_savepoint_after_scenario(self, context, scenario):
        u'''
        Обработчик, срабатывающий после отработки сценария в feature.

        Восстанавливаем данные после прохода сценария, откатившись
        внутри транзакции на заранее сохраненную точку (самое начало
        обработки feature).
        '''
        transaction.savepoint_rollback(context.__initial_savepoint)

    def __init__(self, *args, **kwargs):
        super(RollbackBehaveRunner, self).__init__(*args, **kwargs)

        chosen_rollback_mode = getattr(settings, 'WELLBEHAVED_ROLLBACK_MODE', 'partial')
        # Режим отката изменений в БД
        rollback_hooks = {
            # Режим частичного восстановления
            'partial': {
                'after_scenario': self.restore_savepoint_after_scenario,
                'before_all': self.create_savepoint_before_all
            },
            'manual': {}
        }
        self.hooks.set_wrappers(rollback_hooks[chosen_rollback_mode])
